# Generated by Django 6.0 on 2025-03-06 20:33

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0059_delete_stockrequestproxy"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="container",
            name="max_per_subject",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Maximum number of this container that may be allocated to a subject per stock request. (For example, no more than 3 bottles per subject)",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="max_per_subject",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Maximum number of this container that may be allocated to a subject per stock request. (For example, no more than 3 bottles per subject)",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockrequest",
            name="containers_per_subject",
            field=models.PositiveSmallIntegerField(
                default=3, verbose_name="Number of containers per subject"
            ),
        ),
        migrations.AlterField(
            model_name="stockrequest",
            name="containers_per_subject",
            field=models.PositiveSmallIntegerField(
                default=3, verbose_name="Number of containers per subject"
            ),
        ),
        migrations.CreateModel(
            name="HistoricalStorageBin",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "bin_identifier",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                    ),
                ),
                (
                    "bin_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("name", models.CharField(blank=True, max_length=10, null=True)),
                ("capacity", models.PositiveIntegerField(default=0)),
                ("in_use", models.BooleanField(default=True)),
                (
                    "description",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_pharmacy.location",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Storage bin",
                "verbose_name_plural": "historical Storage bins",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="StorageBin",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "bin_identifier",
                    models.CharField(
                        blank=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "bin_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("name", models.CharField(blank=True, max_length=10, null=True)),
                ("capacity", models.PositiveIntegerField(default=0)),
                ("in_use", models.BooleanField(default=True)),
                (
                    "description",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "location",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.location",
                    ),
                ),
            ],
            options={
                "verbose_name": "Storage bin",
                "verbose_name_plural": "Storage bins",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
            },
        ),
        migrations.CreateModel(
            name="HistoricalStorageBinItem",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "item_identifier",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                    ),
                ),
                (
                    "item_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "stock",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        limit_choices_to={
                            "confirmed_at_site": True,
                            "dispensed": False,
                        },
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_pharmacy.stock",
                    ),
                ),
                (
                    "storage_bin",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_pharmacy.storagebin",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Storage bin item",
                "verbose_name_plural": "historical Storage bin items",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="StorageBinItem",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "item_identifier",
                    models.CharField(
                        blank=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "item_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "stock",
                    models.ForeignKey(
                        limit_choices_to={
                            "confirmed_at_site": True,
                            "dispensed": False,
                        },
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.stock",
                    ),
                ),
                (
                    "storage_bin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.storagebin",
                    ),
                ),
            ],
            options={
                "verbose_name": "Storage bin item",
                "verbose_name_plural": "Storage bin items",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
            },
        ),
        migrations.AddIndex(
            model_name="storagebin",
            index=models.Index(
                fields=["modified", "created"], name="edc_pharmac_modifie_e32a72_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="storagebin",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_pharmac_user_mo_b81f20_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="storagebinitem",
            index=models.Index(
                fields=["modified", "created"], name="edc_pharmac_modifie_46e15a_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="storagebinitem",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_pharmac_user_mo_dc165a_idx",
            ),
        ),
    ]



import pandas as pd


def convert_numbers_to_float64_int64(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.select_dtypes(include=["number"]).columns:
        if pd.api.types.is_float_dtype(df[col]):
            df[col] = df[col].astype("float64")
        elif pd.api.types.is_integer_dtype(df[col]):
            # df[col].replace({pd.NA: np.nan})
            df[col] = df[col].astype("float64")
    # df.replace({pd.NA: np.nan})
    return df



from django.core.exceptions import ValidationError


class InvalidAssignment(Exception):  # noqa: N818
    pass


class RandomizationListImportError(Exception):
    pass


class RandomizationListAlreadyImported(Exception):  # noqa: N818
    pass


class RandomizationListError(Exception):
    pass


class RegistryNotLoaded(Exception):  # noqa: N818
    pass


class NotRegistered(Exception):  # noqa: N818
    pass


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class SiteRandomizerError(Exception):
    pass


class RandomizationListExporterError(Exception):
    pass


class SubjectNotRandomization(Exception):  # noqa: N818
    pass


class InvalidAssignmentDescriptionMap(Exception):  # noqa: N818
    pass


class RandomizationListFileNotFound(Exception):  # noqa: N818
    pass


class RandomizationListNotLoaded(Exception):  # noqa: N818
    pass


class RandomizationError(Exception):
    pass


class AlreadyRandomized(ValidationError):  # noqa: N818
    pass


class AllocationError(Exception):
    pass


class RegisterRandomizerError(Exception):
    pass


class RandomizationListModelError(Exception):
    pass



import contextlib
import copy
import sys
from importlib import import_module

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style
from django.db import transaction
from django.utils.module_loading import module_has_submodule

from .load_list_data import LoadListDataError
from .preload_data import PreloadData


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class AlreadyLoaded(Exception):  # noqa: N818
    pass


class SiteListDataError(Exception):
    pass


def get_autodiscover_enabled():
    return getattr(settings, "EDC_LIST_DATA_ENABLE_AUTODISCOVER", True)


class SiteListData:
    """Load list data from any module named "list_data".

    Called in AppConfig or by management command.
    """

    default_module_prefixes = ("edc_",)
    default_module_name = "list_data"

    def __init__(self, module_name=None):
        self.registry = {}
        self.app_names = []
        self.models = []
        self.module_name = module_name or self.default_module_name

    def initialize(self, module_name=None):
        self.__init__(module_name=module_name)

    def register(self, module, app_name=None):
        """Registers but does NOT `load` list_data."""
        if app_name and app_name in self.app_names:
            raise AlreadyLoaded(f"App already loaded. Got {app_name}.")
        self.app_names.append(app_name)
        opts = {k: v for k, v in self._get_options(module).items()}
        sys.stdout.write(f"   - registered {self.module_name} from '{module.__name__}'\n")
        if opts.get(self.module_name):
            self._replace_list_data_or_raise_on_duplicate(module, opts)
        self.registry[module.__name__] = opts

    def load_data(self) -> None:
        """Calls `load` class with each list_data dictionary module to
        update database `list` tables.

        See post_migrate_list_data.
        """
        style = color_style()
        for module_name, opts in self.registry.items():
            sys.stdout.write(f"   - loading {module_name} ... \r")
            try:
                with transaction.atomic():
                    obj = PreloadData(**opts)
                    sys.stdout.write(
                        f"   - loading {module_name} ... {obj.item_count} items.\n"
                    )
            except LoadListDataError as e:
                sys.stdout.write(style.ERROR(f"ERROR! {e}\n"))

    def _replace_list_data_or_raise_on_duplicate(self, module, opts: dict) -> None:
        """Raises on duplicates or replaces.

        1. Checks `list model` has not been registered from another app
        2. Updates the list of `list models` already registered.
        3. If duplicate is found raises OR replaces.

        Note: * edc_* modules should load first.
              * only edc_* modules can provide defaults
        """
        models = []
        for label_lower in opts.get(self.module_name):
            default_module_name = self._get_default_module_name(module, label_lower)
            if label_lower not in self.models:
                models.append(label_lower)
            elif (
                label_lower in self.models
                and default_module_name
                and module.__name__ != default_module_name
            ):
                self._replace_default_list_data_for_model(
                    label_lower, module.__name__, default_module_name
                )
            else:
                raise AlreadyRegistered(
                    "List data for table is already registered. "
                    f"Got table {label_lower} in {module.__name__}."
                )
        self.models.extend(models)

    def _get_default_module_name(self, module, label_lower: str) -> str:
        for full_module_name, opts in self.registry.items():
            if label_lower in (opts.get(self.module_name) or {}):
                for prefix in self.default_module_prefixes:
                    if (
                        full_module_name.startswith(prefix)
                        and full_module_name != module.__name__
                    ):
                        return full_module_name
        return ""

    def _replace_default_list_data_for_model(
        self, label_lower: str, full_module_name: str, default_module_name: str
    ) -> None:
        self.registry[default_module_name][self.module_name].pop(label_lower)
        sys.stdout.write(
            f"     + {self.module_name} from `{full_module_name}.{label_lower}`\n"
            f"         has replaced `{default_module_name}.{label_lower}`.\n"
        )

    @staticmethod
    def _get_options(module) -> dict:
        opts: dict = {}
        opts.update(list_data=getattr(module, "list_data", None))
        opts.update(model_data=getattr(module, "model_data", None))
        opts.update(list_data_model_name=getattr(module, "list_data_model_name", None))
        opts.update(apps=getattr(module, "apps", None))
        if not any([x for x in opts.values()]):
            raise SiteListDataError(f"Invalid list_data module. See {module}")
        return opts

    def autodiscover(self) -> None:
        if (
            get_autodiscover_enabled()
            and "makemigrations" not in sys.argv
            and "showmigrations" not in sys.argv
        ):
            sys.stdout.write(f" * checking sites for `{self.module_name}` ...\n")
            for app_name in django_apps.app_configs:
                with contextlib.suppress(ModuleNotFoundError):
                    self._import_and_register(app_name)

    def _import_and_register(self, app_name: str) -> None:
        mod = import_module(app_name)
        before_import_registry = copy.deepcopy(site_list_data.registry)
        try:
            module = import_module(f"{app_name}.{self.module_name}")
        except ImportError as e:
            site_list_data.registry = before_import_registry
            if module_has_submodule(mod, self.module_name):
                raise SiteListDataError(f"{e} See {app_name}.{self.module_name}") from e
        else:
            self.register(module, app_name=app_name)


site_list_data = SiteListData()



import os
import pwd

from django.db.models import CharField
from django.utils.translation import gettext as _


class UserField(CharField):
    description = _("Custom field for user created")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        CharField.__init__(self, *args, **kwargs)

    def get_os_username(self):
        return pwd.getpwuid(os.getuid()).pw_name

    def pre_save(self, model_instance, add):
        """Updates username created on ADD only."""
        value = super(UserField, self).pre_save(model_instance, add)
        if not value and not add:
            # fall back to OS user if not accessing through browser
            # better than nothing ...
            value = self.get_os_username()
            setattr(model_instance, self.attname, value)
            return value
        return value



from warnings import warn

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist


class ExportPermissionsFixer:
    """
    Fix permissions when new permissions are added to the default list
    in the model Meta class.

    For example. see `edc_models.BaseUuidModel.Meta`

    Usage:
        fixer = ExportPermissionsFixer(warn_only=True)
        fixer.fix()
    """

    def __init__(self, app_label=None, verbose=None, warn_only=None):
        if app_label:
            self.app_configs = [django_apps.get_app_config(app_label)]
        else:
            self.app_configs = django_apps.get_app_configs()
        self.verbose = verbose if verbose is not None else False
        self.warn_only = warn_only

    def fix(self):
        """Add "import" and "export" to default permissions of add, change, delete, view.

        Needed for BaseUuidModel models with an initial migration created before edc==0.1.24.

        Provided for ``django-import-export`` integration
        """

        if self.verbose:
            print("Adding `import` and `export` to default permissions.")  # noqa: T201
        for app_config in self.app_configs:
            if self.verbose:
                print(f"  * updating {app_config.name}")  # noqa: T201
            for model in app_config.get_models():
                self.fix_for_model(model)
        if self.verbose:
            print("Done")  # noqa: T201

    def fix_for_model(self, model):
        from edc_model import models as edc_models  # noqa: PLC0415

        if issubclass(model, (edc_models.BaseUuidModel,)):
            permission_model_cls = django_apps.get_model("auth.permission")
            content_type_model_cls = django_apps.get_model("contenttypes.contenttype")
            if self.verbose:
                print(f"    - {model._meta.label_lower}")  # noqa: T201
            try:
                app_label, model_name = model._meta.label_lower.split(".")
                content_type = content_type_model_cls.objects.get(
                    app_label=app_label,
                    model__iexact=model_name,
                )
            except ObjectDoesNotExist as e:
                if self.warn_only:
                    warn(f"ObjectDoesNotExist: {e} Got {model}.", stacklevel=2)
                else:
                    raise ObjectDoesNotExist(f"{e} Got {model}.") from e
            else:
                for action in ["import", "export"]:
                    codename = f"{action}_{model._meta.label_lower.split('.')[1]}"
                    opts = dict(content_type=content_type, codename=codename)
                    try:
                        obj = permission_model_cls.objects.get(**opts)
                    except ObjectDoesNotExist:
                        opts.update(name=f"Can {action} {model._meta.verbose_name}")
                        permission_model_cls.objects.create(**opts)
                        if self.verbose:
                            print(f"       created for {model._meta.label_lower}")  # noqa: T201
                    else:
                        obj.name = f"Can {action} {model._meta.verbose_name}"
                        obj.save()



from django.conf import settings

__all__ = ["get_default_reportable_grades"]


def get_default_reportable_grades() -> list[str]:
    return getattr(settings, "EDC_REPORTABLE_DEFAULT_REPORTABLE_GRADES", [3, 4])



from django.contrib import admin
from django.contrib.admin import ModelAdmin
from edc_action_item.modeladmin_mixins import ActionItemModelAdminMixin


class ReportableResultsModelAdminMixin(ActionItemModelAdminMixin, ModelAdmin):
    form = None

    fieldsets = None

    autocomplete_fields = ["requisition"]

    radio_fields = {  # noqa: RUF012
        "results_abnormal": admin.VERTICAL,
        "results_reportable": admin.VERTICAL,
    }

    # TODO: add filter to see below grade 3,4
    def get_list_filter(self, request) -> tuple:
        fields = super().get_list_filter(request)
        custom_fields = ("missing_count", "results_abnormal", "results_reportable")
        return tuple(f for f in custom_fields if f not in fields) + fields

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = list(fields)
        custom_fields = [
            "missing_count",
            "missing",
            "abnormal",
            "reportable",
            "action_identifier",
        ]
        fields[4:1] = custom_fields
        return fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        fields = super().get_readonly_fields(request, obj=obj)
        custom_fields = ("summary", "reportable_summary", "abnormal_summary", "errors")
        return tuple(set(fields + custom_fields))

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = list(fields)
        fields.insert(0, "subject_visit__subject_identifier")
        return tuple(set(fields))

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "appointment" and request.GET.get("appointment"):
            kwargs["queryset"] = db_field.related_model.objects.filter(
                pk=request.GET.get("appointment", 0)
            )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.management import CommandError

from ..exceptions import PrescriptionAlreadyExists, PrescriptionError

if TYPE_CHECKING:
    from ..models import Rx


def create_prescription(
    subject_identifier: str,
    report_datetime: datetime,
    medication_names: list[str],
    randomizer_name: str | None = None,
    site: Any | None = None,
    site_id: Any | None = None,
    apps: Any | None = None,
) -> Rx:
    """Creates a PrescriptionAction and Rx model instance"""
    site_id = site_id or site.id
    medication_model_cls = (apps or django_apps).get_model("edc_pharmacy.medication")
    rx_model_cls = (apps or django_apps).get_model("edc_pharmacy.rx")
    medications = []
    for medication_name in medication_names:
        try:
            obj = medication_model_cls.objects.get(name__iexact=medication_name)
        except ObjectDoesNotExist as e:
            raise PrescriptionError(
                "Unable to create prescription. Medication does not exist. "
                f"Got {medication_name}"
            ) from e
        else:
            medications.append(obj)
    try:
        rx = rx_model_cls.objects.get(subject_identifier=subject_identifier)
    except ObjectDoesNotExist:
        opts = dict(
            subject_identifier=subject_identifier,
            report_datetime=report_datetime,
            rx_date=report_datetime.date(),
            randomizer_name=randomizer_name,
        )
        if site_id:
            opts.update(site_id=site_id)
        try:
            rx = rx_model_cls.objects.create(**opts)
        except ObjectDoesNotExist as e:
            raise CommandError(f"Site does not exists. site_id={site_id}. Got {e}") from e
        for obj in medications:
            rx.medications.add(obj)
    else:
        raise PrescriptionAlreadyExists(f"Prescription already exists. Got {rx}")
    return rx



# Generated by Django 6.0 on 2026-01-14 22:08

import _socket
import django.db.models.deletion
import django.utils.timezone
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_pharmacy.models.stock.confirmation_at_location_item
import edc_sites.managers
import simple_history.models
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0091_alter_allocation_revision_alter_assignment_revision_and_more",
        ),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RenameModel(
            old_name="HistoricalConfirmationAtSiteItem",
            new_name="HistoricalConfirmationAtLocationItem",
        ),
        migrations.RenameModel(
            old_name="ConfirmationAtSiteItem",
            new_name="ConfirmationAtLocationItem",
        ),
        migrations.RenameModel(
            old_name="HistoricalConfirmationAtSite",
            new_name="HistoricalConfirmationAtLocation",
        ),
        migrations.RenameModel(
            old_name="ConfirmationAtSite",
            new_name="ConfirmationAtLocation",
        ),
    ]



# Generated by Django 5.2.10 on 2026-01-24 16:32

import _socket
import django.db.models.deletion
import django.utils.timezone
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0112_historicalstock_container_qty_stock_container_qty"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalstock",
            name="container_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units per container. ",
                max_digits=20,
                null=True,
                verbose_name="Units per container",
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="container_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units per container. ",
                max_digits=20,
                null=True,
                verbose_name="Units per container",
            ),
        ),
        migrations.CreateModel(
            name="StockTransactionType",
            fields=[
                (
                    "name",
                    models.CharField(
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(
                        default="", max_length=250, verbose_name="Plural name"
                    ),
                ),
                (
                    "display_name",
                    models.CharField(
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        default="",
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                    ),
                ),
                ("extra_value", models.CharField(default="", max_length=250)),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "verbose_name": "Stock transaction type",
                "verbose_name_plural": "Stock transaction types",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "indexes": [
                    models.Index(
                        fields=["display_index", "display_name"],
                        name="edc_pharmac_display_46f88a_idx",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="StockTransaction",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        default="",
                        editable=False,
                        help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                        max_length=75,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        default="",
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        default="",
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        default="",
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=10,
                        verbose_name="Device created",
                    ),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=10,
                        verbose_name="Device modified",
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "transaction_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "transaction_qty",
                    models.DecimalField(
                        decimal_places=2, default=Decimal("0.0"), max_digits=20
                    ),
                ),
                (
                    "balance",
                    models.DecimalField(
                        decimal_places=2, default=Decimal("0.0"), max_digits=20
                    ),
                ),
                (
                    "stock",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.stock",
                    ),
                ),
                (
                    "transaction_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.stocktransactiontype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Stock transaction",
                "verbose_name_plural": "Stock transactions",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["modified", "created"],
                        name="edc_pharmac_modifie_69d29d_idx",
                    ),
                    models.Index(
                        fields=["user_modified", "user_created"],
                        name="edc_pharmac_user_mo_01143b_idx",
                    ),
                ],
            },
        ),
    ]



PROTOCOL_VIOLATION = "PROTOCOL_VIOLATION"
PROTOCOL_VIOLATION_VIEW = "PROTOCOL_VIOLATION_VIEW"
PROTOCOL_INCIDENT = "PROTOCOL_INCIDENT"
PROTOCOL_INCIDENT_VIEW = "PROTOCOL_INCIDENT_VIEW"


protocol_violation_codenames = (
    "edc_protocol_incident.add_protocoldeviationviolation",
    "edc_protocol_incident.change_protocoldeviationviolation",
    "edc_protocol_incident.delete_protocoldeviationviolation",
    "edc_protocol_incident.view_protocoldeviationviolation",
)

protocol_violation_view_codenames = ("edc_protocol_incident.view_protocoldeviationviolation",)


protocol_incident_codenames = (
    "edc_protocol_incident.add_protocolincident",
    "edc_protocol_incident.change_protocolincident",
    "edc_protocol_incident.delete_protocolincident",
    "edc_protocol_incident.view_protocolincident",
)

protocol_incident_view_codenames = ("edc_protocol_incident.view_protocolincident",)



# Generated by Django 5.1.2 on 2024-11-15 14:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0026_historicalstockrequest_cutoff_datetime_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalstock",
            old_name="at_location",
            new_name="transferred",
        ),
        migrations.RenameField(
            model_name="stock",
            old_name="at_location",
            new_name="transferred",
        ),
    ]



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from clinicedc_constants import HIV, YES
from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from edc_model.utils import model_exists_or_raise
from edc_visit_schedule.baseline import VisitScheduleBaselineError
from edc_visit_schedule.utils import is_baseline

if TYPE_CHECKING:
    from edc_model.models import BaseUuidModel
    from edc_visit_tracking.model_mixins import VisitTrackingCrfModelMixin

    class CrfLikeModel(VisitTrackingCrfModelMixin, BaseUuidModel):
        pass


EDC_DX_REVIEW_APP_LABEL = getattr(settings, "EDC_DX_REVIEW_APP_LABEL", "edc_dx_review")


class ModelNotDefined(Exception):
    pass


class BaselineModelError(Exception):
    pass


def get_list_model_app():
    return getattr(
        settings, "EDC_DX_REVIEW_LIST_MODEL_APP_LABEL", settings.LIST_MODEL_APP_LABEL
    )


def get_clinical_review_baseline_model_cls():
    clinicalreviewbaseline = get_extra_attrs().get(
        "clinicalreviewbaseline", "clinicalreviewbaseline"
    )

    return django_apps.get_model(f"{EDC_DX_REVIEW_APP_LABEL}.{clinicalreviewbaseline}")


def get_clinical_review_model_cls():
    clinicalreview = get_extra_attrs().get("clinicalreview", "clinicalreview")
    return django_apps.get_model(f"{EDC_DX_REVIEW_APP_LABEL}.{clinicalreview}")


def get_medication_model_cls():
    medications = get_extra_attrs().get("medications", "medications")
    return django_apps.get_model(f"{EDC_DX_REVIEW_APP_LABEL}.{medications}")


def get_initial_review_model_cls(prefix):
    initialreview = get_extra_attrs().get("initialreview", "initialreview")
    return django_apps.get_model(f"{EDC_DX_REVIEW_APP_LABEL}.{prefix.lower()}{initialreview}")


def get_review_model_cls(prefix):
    review = get_extra_attrs().get("review", "review")
    return django_apps.get_model(f"{EDC_DX_REVIEW_APP_LABEL}.{prefix.lower()}{review}")


def get_extra_attrs():
    """Settings from EDC_DX_REVIEW_EXTRA_ATTRS.

    See model name suffixes used in model_cls getters in utils.py.
    """
    extra_attrs = {
        "clinicalreview": "clinicalreview",
        "clinicalreviewbaseline": "clinicalreviewbaseline",
        "initialreview": "initialreview",
        "medications": "medications",
        "review": "review",
    }
    try:
        data = settings.EDC_DX_REVIEW_EXTRA_ATTRS
    except AttributeError:
        pass
    else:
        extra_attrs.update(data)
    return extra_attrs


def raise_if_clinical_review_does_not_exist(subject_visit) -> CrfLikeModel:
    try:
        baseline = is_baseline(instance=subject_visit)
    except VisitScheduleBaselineError as e:
        raise forms.ValidationError(str(e))
    else:
        if baseline:
            model_cls = get_clinical_review_baseline_model_cls()
        else:
            model_cls = get_clinical_review_model_cls()
        try:
            obj = model_exists_or_raise(subject_visit=subject_visit, model_cls=model_cls)
        except ObjectDoesNotExist:
            raise forms.ValidationError(f"Complete {model_cls._meta.verbose_name} CRF first.")
    return obj


def raise_if_both_ago_and_actual_date(
    cleaned_data: dict | None = None,
    date_fld: str | None = None,
    ago_fld: str | None = None,
    dx_ago: str | None = None,
    dx_date: date | None = None,
) -> None:
    if cleaned_data:
        dx_ago = cleaned_data.get(date_fld)
        dx_date = cleaned_data.get(ago_fld)
    if dx_ago and dx_date:
        raise forms.ValidationError(
            {
                ago_fld: (
                    "Date conflict. Do not provide a response "
                    "here if the date of diagnosis is available."
                )
            }
        )


def requires_clinical_review_at_baseline(subject_visit):
    try:
        get_clinical_review_baseline_model_cls().objects.get(
            subject_visit__subject_identifier=subject_visit.subject_identifier
        )
    except ObjectDoesNotExist:
        raise forms.ValidationError(
            "Please complete the "
            f"{get_clinical_review_baseline_model_cls()._meta.verbose_name} first."
        )


def raise_if_initial_review_does_not_exist(subject_visit, prefix):
    model_exists_or_raise(
        subject_visit=subject_visit,
        model_cls=get_initial_review_model_cls(prefix),
    )


def raise_if_review_does_not_exist(subject_visit, prefix):
    model_exists_or_raise(
        subject_visit=subject_visit,
        model_cls=get_review_model_cls(prefix),
    )


def medications_exists_or_raise(subject_visit) -> bool:
    if subject_visit:
        try:
            get_medication_model_cls().objects.get(subject_visit=subject_visit)
        except ObjectDoesNotExist:
            raise forms.ValidationError(
                f"Complete the `{get_medication_model_cls()._meta.verbose_name}` CRF first."
            )
    return True


def is_rx_initiated(
    subject_identifier: str, report_datetime: datetime, instance_id: str | None = None
) -> bool:
    """Return True if already initiated"""
    try:
        get_initial_review_model_cls(HIV).objects.get(
            subject_visit__subject_identifier=subject_identifier,
            report_datetime__lte=report_datetime,
            rx_init=YES,
        )
    except ObjectDoesNotExist:
        exclude = {}
        if instance_id:
            exclude = {"id": instance_id}
        rx_initiated = (
            get_review_model_cls(HIV)
            .objects.filter(
                subject_visit__subject_identifier=subject_identifier,
                report_datetime__lte=report_datetime,
                rx_init=YES,
            )
            .exclude(**exclude)
            .exists()
        )
    else:
        rx_initiated = True
    return rx_initiated


def art_initiation_date(subject_identifier: str, report_datetime: datetime) -> date:
    """Returns date initiated on ART or None by querying
    the HIV Initial Review and then the HIV Review.
    """
    art_date = None
    try:
        initial_review = get_initial_review_model_cls(HIV).objects.get(
            subject_visit__subject_identifier=subject_identifier,
            report_datetime__lte=report_datetime,
        )
    except ObjectDoesNotExist:
        pass
    else:
        if initial_review.arv_initiated == YES:
            art_date = initial_review.best_art_initiation_date
        else:
            for review in (
                get_review_model_cls(HIV)
                .objects.filter(
                    subject_visit__subject_identifier=subject_identifier,
                    report_datetime__lte=report_datetime,
                )
                .order_by("-report_datetime")
            ):
                if review.arv_initiated == YES:
                    art_date = review.arv_initiation_actual_date
                    break
    return art_date



import sys

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand

from edc_visit_schedule.site_visit_schedules import site_visit_schedules


class Command(BaseCommand):
    help = "List email recipients for each registered notification"

    def add_arguments(self, parser):
        parser.add_argument(
            "--delete",
            action="store_true",
            help="Delete invalid OnSchedule model instances",
        )

    def handle(self, *args, **options):  # noqa: ARG002
        allow_delete = False
        if options["delete"]:
            allow_delete = True
        else:
            sys.stdout.write("Checking only\n")
        subject_schedule_history_cls = django_apps.get_model(
            "edc_visit_schedule.subjectschedulehistory"
        )
        for visit_schedule in site_visit_schedules.visit_schedules.values():
            for schedule in visit_schedule.schedules.values():
                try:
                    onschedule_model_cls = schedule.onschedule_model_cls
                except LookupError:
                    pass
                else:
                    for onschedule_obj in onschedule_model_cls.objects.all():
                        try:
                            subject_schedule_history_cls.objects.get(
                                subject_identifier=onschedule_obj.subject_identifier,
                                onschedule_model=onschedule_model_cls._meta.label_lower,
                            )
                        except ObjectDoesNotExist:
                            msg = (
                                f"{onschedule_model_cls._meta.label_lower} for "
                                f"{onschedule_obj.subject_identifier} is invalid."
                            )
                            if allow_delete:
                                msg = f"{msg} deleted.\n"
                                onschedule_obj.delete()
                            sys.stdout.write(msg)



from decimal import Decimal

from clinicedc_constants import NULL_STRING
from django.core.validators import MinValueValidator
from django.db import models
from edc_model.models import BaseUuidModel, HistoricalRecords

from .container_type import ContainerType
from .container_units import ContainerUnits


class Manager(models.Manager):
    use_in_migrations = True


class Container(BaseUuidModel):
    name = models.CharField(max_length=50, unique=True, blank=False)

    display_name = models.CharField(
        max_length=100, unique=True, default=NULL_STRING, blank=False
    )

    container_type = models.ForeignKey(ContainerType, on_delete=models.PROTECT, null=True)

    units = models.ForeignKey(ContainerUnits, on_delete=models.PROTECT, null=True)

    unit_qty_default = models.DecimalField(
        verbose_name="Default qty",
        decimal_places=2,
        max_digits=10,
        null=True,
        blank=False,
        validators=[MinValueValidator(Decimal("1.0"))],
        help_text="May be adjusted. Default value for entry form",
    )

    unit_qty_places = models.IntegerField(
        default=0.0,
        blank=False,
        validators=[MinValueValidator(0)],
        help_text="May NOT be adjusted at receiving",
    )

    unit_qty_max = models.DecimalField(
        decimal_places=2,
        max_digits=10,
        null=True,
        blank=False,
        validators=[MinValueValidator(Decimal("1.0"))],
        help_text="Maximum unit capacity of container",
    )

    max_items_per_subject = models.IntegerField(
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        help_text=(
            "Maximum number of this container that may be "
            "allocated to a subject per stock request. "
            "(For example, no more than 3 bottles per subject)"
        ),
    )

    may_order_as = models.BooleanField(
        verbose_name="Container may be used for ordering", default=False
    )

    may_receive_as = models.BooleanField(
        verbose_name="Container may be used for receiving", default=False
    )

    may_repack_as = models.BooleanField(
        verbose_name="Container may be used for repack request", default=False
    )

    may_request_as = models.BooleanField(
        verbose_name="Container may be used for stock request",
        default=False,
    )

    may_dispense_as = models.BooleanField(
        verbose_name="Container may be used for dispensing to subject", default=False
    )

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.display_name or self.name

    def save(self, *args, **kwargs):
        if self.may_request_as or self.may_repack_as:
            self.may_order_as = False
            self.may_receive_as = False
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Container"
        verbose_name_plural = "Containers"



from clinicedc_constants import NULL_STRING
from django.db import models
from django.utils import timezone
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value


class Manager(models.Manager):
    use_in_migrations = True


class Confirmation(BaseUuidModel):
    """Track confirmed stock items.

    Confirmed stock items are items created, labels printed, and
    `confirmed` when the label is scanned back into the EDC.

    Only confirmed stock truly exist.
    """

    confirmation_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    stock = models.OneToOneField(
        "edc_pharmacy.stock",
        on_delete=models.PROTECT,
        null=True,
    )

    code = models.CharField(
        verbose_name="Stock code",
        max_length=15,
        default=NULL_STRING,
        blank=True,
        editable=False,
    )

    confirmed_datetime = models.DateTimeField(default=timezone.now)

    confirmed_by = models.CharField(max_length=100, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.confirmation_identifier

    def save(self, *args, **kwargs):
        if not self.confirmation_identifier:
            self.confirmation_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        self.code = self.stock.code
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock Confirmation"
        verbose_name_plural = "Stock Confirmations"



from typing import Any

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.utils.html import escape
from django.views.generic.base import ContextMixin

from edc_lab.models import Aliquot, Box, BoxItem


class BoxViewError(Exception):
    pass


class BoxViewMixin(ContextMixin):
    """Declare with the ModelsViewMixin."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._box = None
        self._box_item = None
        self._box_identifier = None
        self._box_item_identifier = None
        self.original_box_item_identifier = None
        self.original_box_identifier = None

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            {
                "box_identifier": self.original_box_identifier,
                "box_item_identifier": self.original_box_item_identifier,
                "box": self.box,
                "paginator_url_kwargs": self.url_kwargs,
            }
        )
        return super().get_context_data(**kwargs)

    @property
    def box_identifier(self):
        """Returns a cleaned box identifier."""
        if not self._box_identifier:
            self.original_box_identifier = escape(self.kwargs.get("box_identifier")).strip()
            self._box_identifier = "".join(self.original_box_identifier.split("-"))
        return self._box_identifier

    @property
    def box_item_identifier(self):
        """Returns a cleaned box_item_identifier or None."""
        if not self._box_item_identifier:
            self.original_box_item_identifier = escape(
                self.request.POST.get("box_item_identifier", "")
            ).strip()
            if self.original_box_item_identifier:
                self._box_item_identifier = self._clean_box_item_identifier()
        return self._box_item_identifier

    @property
    def box(self):
        if not self._box:
            if self.box_identifier:
                try:
                    self._box = Box.objects.get(box_identifier=self.box_identifier)
                except ObjectDoesNotExist:
                    self._box = None
        return self._box

    @property
    def box_item(self):
        """Returns a box item model instance."""
        if not self._box_item:
            if self.box_item_identifier:
                try:
                    self._box_item = BoxItem.objects.get(
                        box=self.box, identifier=self.box_item_identifier
                    )
                except ObjectDoesNotExist:
                    message = f"Invalid box item. Got {self.original_box_item_identifier}"
                    messages.error(self.request, message)
        return self._box_item

    def get_box_item(self, position):
        """Returns a box item model instance for the given position."""
        try:
            box_item = BoxItem.objects.get(box=self.box, position=position)
        except ObjectDoesNotExist:
            message = f"Invalid position for box. Got {position}"
            messages.error(self.request, message)
            return None
        return box_item

    def _clean_box_item_identifier(self):
        """Returns a valid identifier or raises."""
        box_item_identifier = "".join(self.original_box_item_identifier.split("-"))
        try:
            obj = Aliquot.objects.get(aliquot_identifier=box_item_identifier)
        except ObjectDoesNotExist:
            message = "Invalid aliquot identifier. Got {}.".format(
                self.original_box_item_identifier or "None"
            )
            messages.error(self.request, message)
            # raise BoxViewError(message)
        else:
            if obj.is_primary and not self.box.accept_primary:
                message = f'Box does not accept "primary" specimens. Got {self.original_box_item_identifier} is primary.'
                messages.error(self.request, message)
                # raise BoxViewError(message)
            elif obj.numeric_code not in self.box.specimen_types.split(","):
                message = (
                    "Invalid specimen type. Box accepts types {}. Got {} is type {}.".format(
                        ", ".join(self.box.specimen_types.split(",")),
                        self.original_box_item_identifier,
                        obj.numeric_code,
                    )
                )
                messages.error(self.request, message)
                # raise BoxViewError(message)
        return box_item_identifier



# Generated by Django 6.0 on 2025-02-27 08:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0054_historicalreceiveitem_task_id_receiveitem_task_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalreceiveitem",
            name="comment",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historicalreceiveitem",
            name="reference",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="receiveitem",
            name="comment",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="receiveitem",
            name="reference",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
    ]



from django.contrib import messages

from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab.labels import AliquotLabel
from edc_label import LabPrintersMixin, add_job_results_to_messages

from .action_view import ActionView


class AliquotView(EdcViewMixin, LabPrintersMixin, ActionView):
    post_action_url = "aliquot_listboard_url"
    valid_form_actions = ("print_labels",)
    action_name = "aliquot"
    label_cls = AliquotLabel

    def process_form_action(self, request=None):
        if self.action == "print_labels":
            if not self.selected_items:
                message = "Nothing to do. No items have been selected."
                messages.warning(request, message)
            else:
                job_result = self.print_labels(pks=self.selected_items, request=request)
                if job_result:
                    add_job_results_to_messages(request, [job_result])
                else:
                    messages.error(
                        request,
                        f"Failed to print. Selected items were {self.selected_items}.",
                    )



from __future__ import annotations

from typing import TYPE_CHECKING

from clinicedc_constants import YES
from django.apps import apps as django_apps
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_dashboard.url_names import url_names

from .requisition_listboard_view import RequisitionListboardView

if TYPE_CHECKING:
    from django.db.models import Q

app_config = django_apps.get_app_config("edc_lab_dashboard")


class ReceiveListboardView(RequisitionListboardView):
    action_name = "receive"
    form_action_url = "receive_form_action_url"  # url_name
    listboard_template = "receive_listboard_template"
    listboard_url = "receive_listboard_url"  # url_name
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_receive_listboard"
    listboard_view_only_my_permission_codename = None
    navbar_selected_item = "receive"
    process_listboard_url = "process_listboard_url"  # url_name
    show_all = True
    search_form_url = "receive_listboard_url"  # url_name

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update(is_drawn=YES, clinic_verified=YES, received=False, processed=False)
        return q_object, options

    def get_empty_queryset_message(self) -> str:
        href = reverse(url_names.get(self.process_listboard_url))
        return format_html(
            "All specimens have been received. Continue to "
            '<a href="{}" class="alert-link">processing</a>',
            mark_safe(href),  # nosec B703, B308
        )



# Generated by Django 4.2.7 on 2023-12-03 00:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_export", "0016_alter_datarequest_device_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="datarequest",
            options={},
        ),
        migrations.AlterModelOptions(
            name="datarequesthistory",
            options={
                "verbose_name": "Data Request History",
                "verbose_name_plural": "Data Request History",
            },
        ),
        migrations.AlterModelOptions(
            name="exportreceipt",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="filehistory",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="historicalobjecthistory",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Object History",
                "verbose_name_plural": "historical Object Histories",
            },
        ),
        migrations.AlterModelOptions(
            name="objecthistory",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Object History",
                "verbose_name_plural": "Object Histories",
            },
        ),
        migrations.AlterModelOptions(
            name="plan",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="uploadexportreceiptfile",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AddField(
            model_name="datarequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="datarequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="datarequesthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="datarequesthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="exportreceipt",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="exportreceipt",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="filehistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="filehistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicaldatarequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicaldatarequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalexportreceipt",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalexportreceipt",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalfilehistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalfilehistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalobjecthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalobjecthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalplan",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalplan",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="objecthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="objecthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="plan",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="plan",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="uploadexportreceiptfile",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="uploadexportreceiptfile",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddIndex(
            model_name="datarequest",
            index=models.Index(fields=["name"], name="edc_export__name_443690_idx"),
        ),
        migrations.AddIndex(
            model_name="datarequesthistory",
            index=models.Index(
                fields=["-exported_datetime"], name="edc_export__exporte_ba8050_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="exportreceipt",
            index=models.Index(fields=["-timestamp"], name="edc_export__timesta_6a2406_idx"),
        ),
        migrations.AddIndex(
            model_name="filehistory",
            index=models.Index(
                fields=["-sent_datetime"], name="edc_export__sent_da_01059e_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="objecthistory",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_export__modifie_36e5ad_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="objecthistory",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_export__user_mo_baeb31_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="plan",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_export__modifie_db36ec_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="plan",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_export__user_mo_659169_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="uploadexportreceiptfile",
            index=models.Index(fields=["-created"], name="edc_export__created_f54b3c_idx"),
        ),
    ]



from clinicedc_constants import DEAD, OTHER

from edc_consent.constants import CONSENT_WITHDRAWAL
from edc_visit_tracking.constants import COMPLETED_PROTOCOL_VISIT, LOST_VISIT

OFFSTUDY_REASONS = (
    (LOST_VISIT, "Lost to follow-up"),
    (COMPLETED_PROTOCOL_VISIT, "Completed protocol"),
    (CONSENT_WITHDRAWAL, "Consent withdrawn"),
    (DEAD, "Deceased"),
    (OTHER, "Other, please specify ..."),
)



from __future__ import annotations

from typing import TYPE_CHECKING

from .constants import IN_PROGRESS_APPT
from .utils import update_appt_status

if TYPE_CHECKING:
    from .models import Appointment


class AppointmentStatusUpdaterError(Exception):
    pass


class AppointmentStatusUpdater:
    def __init__(
        self,
        appointment: Appointment,
        change_to_in_progress: bool | None = None,
        clear_others_in_progress: bool | None = None,
    ):
        self.appointment = appointment
        if "historical" in self.appointment._meta.label_lower:
            raise AppointmentStatusUpdaterError(
                f"Not an Appointment model instance. Got {self.appointment._meta.label_lower}."
            )
        if not getattr(self.appointment, "id", None):
            raise AppointmentStatusUpdaterError(
                "Appointment instance must exist. Got `id` is None"
            )
        if change_to_in_progress and self.appointment.appt_status != IN_PROGRESS_APPT:
            self.appointment.appt_status = IN_PROGRESS_APPT
            self.appointment.save_base(update_fields=["appt_status"])
        if clear_others_in_progress:
            for appt in self.appointment.__class__.objects.filter(
                visit_schedule_name=self.appointment.visit_schedule_name,
                schedule_name=self.appointment.schedule_name,
                appt_status=IN_PROGRESS_APPT,
            ).exclude(id=self.appointment.id):
                update_appt_status(appt, save=True)



# Generated by Django 3.0.4 on 2020-05-11 22:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_export", "0004_auto_20190305_0123"),
    ]

    operations = [
        migrations.CreateModel(
            name="ExportData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("display_export_admin_action", "Display export action"),),
            },
        ),
        migrations.CreateModel(
            name="ImportData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("show_import_admin_action", "Show import action"),),
            },
        ),
    ]



from .assignment import Assignment
from .dosage_guideline import DosageGuideline
from .formulation import Formulation
from .formulation_type import FormulationType
from .frequency_units import FrequencyUnits
from .medication import Medication
from .route import Route
from .units import Units



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from django import forms
from django.db.models import Sum

from ...models import ReceiveItem, Stock

if TYPE_CHECKING:
    from ...models import Container


class ReceiveItemForm(forms.ModelForm):
    @property
    def container(self) -> Container | None:
        return self.cleaned_data.get("container") or self.instance.container

    @property
    def unit_qty_received(self) -> Decimal | None:
        if self.cleaned_data.get("container_unit_qty") is not None and self.cleaned_data.get(
            "item_qty_received"
        ):
            return self.cleaned_data.get("container_unit_qty") * self.cleaned_data.get(
                "item_qty_received"
            )
        return None

    @property
    def total_unit_qty_received(self) -> Decimal:
        """Returns the total received so far for this order item
        not including this transaction.
        """
        return self._meta.model.objects.filter(order_item=self.order_item).exclude(
            id=self.instance.id
        ).aggregate(unit_qty=Sum("unit_qty_received"))["unit_qty"] or Decimal("0.0")

    @property
    def unit_qty_added_to_stock(self) -> Decimal:
        """Returns the unit qty already added to stock from this receive_item"""
        if self.instance.id:
            return Stock.objects.filter(receive_item=self.instance).aggregate(
                unit_qty=Sum("unit_qty_in")
            )["unit_qty"] or Decimal("0.0")
        return Decimal("0.0")

    def clean(self):
        cleaned_data = super().clean()
        if not self.instance.id:
            if not cleaned_data.get("receive"):
                raise forms.ValidationError({"receive": "This field is required"})
            if not cleaned_data.get("order_item"):
                raise forms.ValidationError({"order_item": "This field is required"})
            if not cleaned_data.get("batch"):
                raise forms.ValidationError({"batch": "This field is required"})
            if not cleaned_data.get("reference"):
                raise forms.ValidationError({"reference": "This field is required"})
            if not cleaned_data.get("container"):
                raise forms.ValidationError({"container": "This field is required"})

        if self.container and not self.container.may_receive_as:
            raise forms.ValidationError(
                f"Container is not configured for receiving. See {self.container}."
            )

        if (
            cleaned_data.get("order_item")
            and cleaned_data.get("lot")
            and (
                cleaned_data.get("order_item").product.assignment
                != cleaned_data.get("lot").assignment
            )
        ):
            raise forms.ValidationError({"lot": "Lot assignment does not match product"})

        if self.unit_qty_received and (
            self.unit_qty_received + self.total_unit_qty_received
            > self.order_item.unit_qty_ordered
        ):
            raise forms.ValidationError(
                {
                    "item_qty_received": "Unit quantity received will be greater "
                    "than the unit quantity ordered. "
                    f"Got {self.unit_qty_received + self.total_unit_qty_received} received > "
                    f"{self.order_item.unit_qty_ordered} ordered."
                }
            )
        if self.unit_qty_added_to_stock > self.unit_qty_received:
            raise forms.ValidationError(
                {
                    "item_qty_received": "Unit quantity already added to stock exceeds "
                    "for this amount."
                    f"Got {self.unit_qty_received} received > "
                    f"{self.order_item.unit_qty_ordered} ordered."
                }
            )
        return cleaned_data

    @property
    def order_item(self):
        return self.cleaned_data.get("order_item") or self.instance.order_item

    class Meta:
        model = ReceiveItem
        fields = "__all__"



# Generated by Django 5.0.7 on 2024-08-03 19:55

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django.utils.timezone
import django_revision.revision_field
from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    replaces = [
        ("edc_auth", "0001_initial"),
        ("edc_auth", "0002_auto_20180803_0515"),
        ("edc_auth", "0003_auto_20180803_0532"),
        ("edc_auth", "0004_auto_20180804_2041"),
        ("edc_auth", "0005_userprofile_job_title"),
        ("edc_auth", "0006_userprofile_notifications"),
        ("edc_auth", "0007_auto_20181025_0150"),
        ("edc_auth", "0008_auto_20191022_0134"),
        ("edc_auth", "0009_userprofile_alternate_email"),
        ("edc_auth", "0010_auto_20191022_0422"),
        ("edc_auth", "0011_auto_20191025_2335"),
        ("edc_auth", "0012_auto_20191026_0034"),
        ("edc_auth", "0013_auto_20191026_0055"),
        ("edc_auth", "0014_auto_20191026_1841"),
        ("edc_auth", "0015_auto_20191026_2149"),
        ("edc_auth", "0016_auto_20191026_2153"),
        ("edc_auth", "0017_auto_20191026_2231"),
        ("edc_auth", "0018_auto_20191029_2039"),
        ("edc_auth", "0019_userprofile_export_format"),
        ("edc_auth", "0020_auto_20210120_0005"),
        ("edc_auth", "0021_auto_20210125_0421"),
        ("edc_auth", "0022_auto_20210125_2214"),
        ("edc_auth", "0023_auto_20210423_1451"),
        ("edc_auth", "0024_alter_userprofile_export_format"),
        ("edc_auth", "0025_permissions"),
        ("edc_auth", "0026_rename_permissions_edcpermissions"),
        ("edc_auth", "0027_alter_edcpermissions_options"),
        ("edc_auth", "0028_alter_role_options"),
        ("edc_auth", "0029_alter_edcpermissions_device_created_and_more"),
        ("edc_auth", "0030_alter_role_options_and_more"),
        ("edc_auth", "0031_alter_edcpermissions_options_and_more"),
        ("edc_auth", "0032_userprofile_is_multisite_viewer"),
        ("edc_auth", "0033_alter_userprofile_is_multisite_viewer"),
    ]

    initial = True

    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
        ("edc_notification", "0002_update_notifications"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Role",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "short_name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
                ("groups", models.ManyToManyField(to="auth.group")),
            ],
            options={
                "ordering": ["display_index", "name"],
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="UserProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "clinic_label_printer",
                    models.CharField(
                        blank=True,
                        help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "lab_label_printer",
                    models.CharField(
                        blank=True,
                        help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "print_server",
                    models.CharField(
                        blank=True,
                        help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("sites", models.ManyToManyField(blank=True, to="sites.site")),
                ("job_title", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "email_notifications",
                    models.ManyToManyField(
                        blank=True,
                        limit_choices_to={"enabled": True},
                        related_name="email_notifications",
                        to="edc_notification.notification",
                    ),
                ),
                (
                    "mobile",
                    models.CharField(
                        blank=True,
                        help_text="e.g. +1234567890",
                        max_length=25,
                        null=True,
                        validators=[django.core.validators.RegexValidator(regex="^\\+\\d+")],
                    ),
                ),
                (
                    "sms_notifications",
                    models.ManyToManyField(
                        blank=True,
                        limit_choices_to={"enabled": True},
                        related_name="sms_notifications",
                        to="edc_notification.notification",
                    ),
                ),
                (
                    "alternate_email",
                    models.EmailField(
                        blank=True,
                        max_length=254,
                        null=True,
                        verbose_name="Alternate email address",
                    ),
                ),
                ("user_roles", models.ManyToManyField(to="edc_auth.role")),
            ],
        ),
        migrations.AddIndex(
            model_name="role",
            index=models.Index(
                fields=["id", "name", "display_index"], name="edc_auth_ro_id_587a9b_idx"
            ),
        ),
        migrations.RenameField(
            model_name="userprofile",
            old_name="user_roles",
            new_name="roles",
        ),
        migrations.AlterModelOptions(
            name="role",
            options={"ordering": ["display_index", "display_name"]},
        ),
        migrations.RemoveIndex(
            model_name="role",
            name="edc_auth_ro_id_587a9b_idx",
        ),
        migrations.RemoveField(
            model_name="role",
            name="field_name",
        ),
        migrations.RemoveField(
            model_name="role",
            name="version",
        ),
        migrations.RenameField(
            model_name="role",
            old_name="name",
            new_name="display_name",
        ),
        migrations.AddIndex(
            model_name="role",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_auth_ro_id_cc6bf4_idx",
            ),
        ),
        migrations.RenameField(
            model_name="role",
            old_name="short_name",
            new_name="name",
        ),
        migrations.AlterField(
            model_name="role",
            name="display_name",
            field=models.CharField(
                db_index=True,
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Display Name",
            ),
        ),
        migrations.AlterField(
            model_name="role",
            name="name",
            field=models.CharField(
                db_index=True,
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="roles",
            field=models.ManyToManyField(blank=True, to="edc_auth.role"),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="id",
            field=models.AutoField(primary_key=True, serialize=False),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="export_format",
            field=models.CharField(
                blank=True,
                choices=[
                    ("CSV", "CSV (delimited by pipe `|`)"),
                    (114, "Stata v10 or later"),
                    (117, "Stata v13 or later"),
                    (118, "Stata v14 or later"),
                    (119, "Stata v15 or later"),
                ],
                default="CSV",
                help_text="Note: requires export permissions",
                max_length=25,
                null=True,
                verbose_name="Export format",
            ),
        ),
        migrations.AlterModelOptions(
            name="role",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ["display_index", "display_name"],
            },
        ),
        migrations.CreateModel(
            name="EdcPermissions",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "Edc Permissions",
                "verbose_name_plural": "Edc Permissions",
            },
        ),
        migrations.AlterField(
            model_name="role",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="role",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="role",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="role",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterModelOptions(
            name="role",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.RemoveIndex(
            model_name="role",
            name="edc_auth_ro_id_cc6bf4_idx",
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="role",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="role",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddIndex(
            model_name="role",
            index=models.Index(
                fields=["display_name", "display_index"],
                name="edc_auth_ro_display_0b19e6_idx",
            ),
        ),
        migrations.AlterModelOptions(
            name="edcpermissions",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Edc Permissions",
                "verbose_name_plural": "Edc Permissions",
            },
        ),
        migrations.AlterModelOptions(
            name="userprofile",
            options={
                "verbose_name": "User profile",
                "verbose_name_plural": "User profiles",
            },
        ),
        migrations.AddIndex(
            model_name="edcpermissions",
            index=models.Index(
                fields=["modified", "created"], name="edc_auth_ed_modifie_cd58ed_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="edcpermissions",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_auth_ed_user_mo_817789_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="role",
            index=models.Index(
                fields=["modified", "created"], name="edc_auth_ro_modifie_fe8196_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="role",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_auth_ro_user_mo_dc165e_idx",
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="is_multisite_viewer",
            field=models.BooleanField(
                default=False,
                help_text="Designates if this user may view data from sites other than the current.",
                verbose_name="Multisite viewer",
            ),
        ),
    ]



from clinicedc_constants import NULL_STRING
from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from ...constants import CENTRAL_LOCATION
from ...exceptions import ConfirmAtLocationError
from .confirmation_at_location import ConfirmationAtLocation
from .stock import Stock
from .stock_transfer_item import StockTransferItem


class Manager(models.Manager):
    use_in_migrations = True


class ConfirmationAtLocationItem(SiteModelMixin, BaseUuidModel):
    confirm_at_location = models.ForeignKey(ConfirmationAtLocation, on_delete=models.PROTECT)

    transfer_confirmation_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    transfer_confirmation_item_datetime = models.DateTimeField(default=timezone.now)

    stock = models.OneToOneField(Stock, on_delete=models.PROTECT, null=True)

    code = models.CharField(
        verbose_name="Stock code",
        max_length=15,
        default=NULL_STRING,
        blank=True,
        editable=False,
    )

    location_name = models.CharField(
        verbose_name="Location name",
        max_length=100,
        default=NULL_STRING,
        blank=True,
        editable=False,
    )

    stock_transfer_item = models.OneToOneField(
        StockTransferItem, on_delete=models.PROTECT, null=True
    )

    confirmed_datetime = models.DateTimeField(null=True, blank=True)

    confirmed_by = models.CharField(max_length=150, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.transfer_confirmation_item_identifier} {self.code}"

    def save(self, *args, **kwargs):
        self.code = self.stock_transfer_item.stock.code
        self.location_name = self.confirm_at_location.location.name
        self.site = self.confirm_at_location.site
        if not self.stock_transfer_item:
            raise ConfirmAtLocationError(
                "Field Stock transfer item may not be null. "
                f"Got {self.stock_transfer_item.stock.code}."
            )

        if not self.transfer_confirmation_item_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.transfer_confirmation_item_identifier = f"{next_id:010d}"
        if not (
            (
                self.confirm_at_location.location.site
                == self.stock_transfer_item.stock.allocation.registered_subject.site
            )
            or (self.confirm_at_location.location.name == CENTRAL_LOCATION)
        ):
            raise ConfirmAtLocationError(
                "Location mismatch. Cannot confirm stock item at this location. "
                f"Got {self.stock_transfer_item.stock.code}."
            )
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock confirmation item at location"
        verbose_name_plural = "Stock confirmation items at location"



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse

from ...admin_site import edc_pharmacy_admin
from ...constants import CENTRAL_LOCATION
from ...models import StorageBinItemProxy
from .storage_bin_item_admin import StorageBinItemAdmin


@admin.register(StorageBinItemProxy, site=edc_pharmacy_admin)
class StorageBinItemProxyAdmin(StorageBinItemAdmin):
    change_list_title = "Pharmacy: Central Storage Bin Items"
    change_form_title = "Pharmacy: Central Storage Bin Item"

    list_display = (
        "item_identifier",
        "storage_bin_changelist",
        "stock_changelist",
        "stock__container",
        "subject",
        "location",
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(storage_bin__location__name=CENTRAL_LOCATION)

    @admin.display(description="Bin", ordering="storage_bin__name")
    def storage_bin_changelist(self, obj):
        bin_ref = obj.storage_bin.name or obj.storage_bin.bin_identifier
        url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebinproxy_changelist")
        url = f"{url}?q={bin_ref}"
        context = dict(url=url, label=bin_ref, title="Go to bin")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from typing import Any

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView

from edc_navbar import NavbarViewMixin

from ..view_mixins import EdcViewMixin


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_dashboard/home.html"
    navbar_name = "edc_dashboard"
    navbar_selected_item = "edc_dashboard"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            edc_packages=["not available"],
            third_party_packages=["not available"],
            installed_apps=settings.INSTALLED_APPS,
        )
        return super().get_context_data(**kwargs)

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



# Generated by Django 6.0 on 2025-03-14 12:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0065_allocationproxy"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstock",
            name="stored_at_site",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stock",
            name="stored_at_site",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 5.1.2 on 2024-11-18 01:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0039_remove_dispense_registered_subject_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstock",
            name="dispensed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stock",
            name="dispensed",
            field=models.BooleanField(default=False),
        ),
    ]



from django.db.models import CharField
from django.utils.translation import gettext as _


class OtherCharField(CharField):
    """Field for "Other specify" options."""

    description = _("Custom field for 'Other specify' form field")

    DEFAULT_MAX_LENGTH = 35

    def __init__(self, metadata=None, *args, **kwargs):
        self.metadata = metadata
        kwargs.update(blank=True)
        kwargs.update(default="")
        kwargs.setdefault("max_length", self.DEFAULT_MAX_LENGTH)
        kwargs.setdefault("verbose_name", _("If other, please specify ..."))
        self.max_length = kwargs["max_length"]
        self.verbose_name = kwargs["verbose_name"]
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "CharField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs.update(blank=True)
        kwargs.update(default="")
        kwargs.update(max_length=self.max_length)
        kwargs.update(verbose_name=self.verbose_name)
        if self.metadata is not None:
            kwargs["metadata"] = self.metadata
        return name, path, args, kwargs



from .navbar import Navbar
from .navbar_item import NavbarItem
from .site_navbars import site_navbars
from .utils import get_default_navbar_name, get_register_default_navbar

if get_register_default_navbar():
    default_navbar = Navbar(name=get_default_navbar_name())

    default_navbar.register(
        NavbarItem(
            name="home",
            title="Home",
            fa_icon="fa-home",
            url_without_namespace="home_url",
            codename="edc_navbar.nav_home",
        )
    )

    default_navbar.register(
        NavbarItem(
            name="administration",
            title="Administration",
            fa_icon="fa-cog",
            codename="edc_navbar.nav_administration",
            # url_names_key="administration",
            url_without_namespace="administration_url",
        )
    )

    site_navbars.register(default_navbar)



from .ae_followup import AeFollowupModelMixin
from .ae_initial import (
    AeInitialModelMixin,
    AeInitialSaeModelMixin,
    AeInitialSusarModelMixin,
    AeInitialTmgModelMixin,
)
from .ae_special_interest import (
    AesiFieldsModelMixin,
    AesiMethodsModelMixin,
    AesiModelMixin,
)
from .ae_susar import (
    AeSusarFieldsModelMixin,
    AeSusarMethodsModelMixin,
    AeSusarModelMixin,
)
from .ae_tmg import AeTmgFieldsModelMixin, AeTmgMethodsModelMixin, AeTmgModelMixin
from .death_report import (
    DeathReportExtraFieldsModelMixin,
    DeathReportModelMixin,
    DeathReportTmgModelMixin,
    DeathReportTmgSecondManager,
    DeathReportTmgSecondModelMixin,
    DeathReportTmgSecondSiteManager,
    SimpleDeathReportModelMixin,
)
from .hospitaization import HospitalizationModelMixin



# Generated by Django 5.2.10 on 2026-01-25 22:46

import django.db.models.manager
import edc_sites.managers
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0130_historicalstocktransfer_cancel_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="StorageBinItemProxy",
            fields=[],
            options={
                "verbose_name": "Storage bin item",
                "verbose_name_plural": "Storage bin items",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_pharmacy.storagebinitem",),
            managers=[
                ("on_site", edc_sites.managers.CurrentSiteManager()),
                ("objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name="StorageBinProxy",
            fields=[],
            options={
                "verbose_name": "Storage bin",
                "verbose_name_plural": "Storage bins",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_pharmacy.storagebin",),
            managers=[
                ("on_site", edc_sites.managers.CurrentSiteManager()),
                ("objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



from django.db.models import UniqueConstraint
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel, HistoricalRecords, default_permissions

from ..model_mixins import NoteModelMixin


class Note(NonUniqueSubjectIdentifierFieldMixin, NoteModelMixin):
    """A model class to capture user / dm notes linked to a data query
    report, such as, unmanaged views.

    Unique constraint is on subject_identifier and the report model.

    See also, NoteModelAdminMixin
    """

    history = HistoricalRecords()

    def __str__(self) -> str:
        return f"{self._meta.verbose_name}: {self.subject_identifier}"

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Note"
        verbose_name_plural = "Notes"
        constraints = (
            UniqueConstraint(
                fields=["report_model", "subject_identifier"],
                name="%(app_label)s_%(class)s_report_model_subj_uniq",
            ),
        )
        indexes = BaseUuidModel.Meta.indexes
        default_permissions = *default_permissions, "viewallsites"



from django import forms

from ...models import Lot


class LotForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        if (
            cleaned_data.get("product")
            and cleaned_data.get("assignment")
            and cleaned_data.get("product").assignment != cleaned_data.get("assignment")
        ):
            raise forms.ValidationError({"assignment": "Assignment does not match product"})

        return cleaned_data

    class Meta:
        model = Lot
        fields = "__all__"



# Generated by Django 5.1.2 on 2024-11-20 00:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0047_historicalstockrequest_cancel_stockrequest_cancel_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalorderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(decimal_places=2, max_digits=20, null=True),
        ),
        migrations.AddField(
            model_name="orderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(decimal_places=2, max_digits=20, null=True),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="unit qty ordered less unit qty received",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockrequest",
            name="cancel",
            field=models.CharField(
                blank=True,
                default=None,
                max_length=6,
                null=True,
                verbose_name="To cancel this request, type the word 'CANCEL' here and save the form:",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="unit qty ordered less unit qty received",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="stockrequest",
            name="cancel",
            field=models.CharField(
                blank=True,
                default=None,
                max_length=6,
                null=True,
                verbose_name="To cancel this request, type the word 'CANCEL' here and save the form:",
            ),
        ),
    ]



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from django.db import models
from django.db.models import DecimalField, ExpressionWrapper, F, QuerySet

from ...constants import AVAILABLE
from ...exceptions import InsufficientStockError
from ..medication import Assignment
from .container import Container
from .location import Location

if TYPE_CHECKING:
    from .stock import Stock


class StockManager(models.Manager):
    use_in_migrations = True

    def in_stock(
        self,
        unit_qty: Decimal,
        container: Container,
        location: Location,
        assignment: Assignment,
    ) -> QuerySet[Stock] | None:
        expression_wrapper = ExpressionWrapper(
            F("unit_qty_in") - F("unit_qty_out"),
            output_field=DecimalField(),
        )
        qs = (
            self.get_queryset()
            .filter(
                container=container,
                location=location,
                product__assignment=assignment,
                status=AVAILABLE,
            )
            .annotate(difference=expression_wrapper)
            .filter(difference__gte=unit_qty)
            .order_by("difference")
        )
        if qs.count() == 0:
            raise InsufficientStockError(
                f"Insufficient stock. Got container={container}, "
                f"location={location}, assignment={assignment}"
            )
        return qs[0]



from django.core.exceptions import ValidationError


class SubjectDoesNotExist(Exception):
    pass


class ActionClassNotDefined(Exception):
    pass


class ActionItemError(Exception):
    pass


class ActionItemStatusError(Exception):
    pass


class ActionTypeError(Exception):
    pass


class SingletonActionItemError(Exception):
    pass


class ActionError(ValidationError):
    pass



from __future__ import annotations

import sys
from pathlib import Path

from django.core.management import CommandError, color_style
from django.core.management.base import BaseCommand

from edc_export.constants import CSV, STATA_14
from edc_export.models_to_file import ModelsToFile
from edc_export.utils import (
    get_default_models_for_export,
    get_export_user,
    get_model_names_for_export,
    get_site_ids_for_export,
    validate_user_perms_or_raise,
)
from edc_sites.site import sites as site_sites

ALL_COUNTRIES = "all"

style = color_style()


class Command(BaseCommand):
    def __init__(self, **kwargs):
        self._countries: list[str] = []
        self.options = {}
        self.decrypt: bool | None = None
        self.site_ids: list[int] = []
        self.exclude_historical: bool | None = None
        super().__init__(**kwargs)

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--app",
            dest="app_labels",
            default="",
            help="app label. Separate by comma if more than one.",
        )

        parser.add_argument(
            "-m",
            "--model",
            dest="model_names",
            default="",
            help="model name in label_lower format. Separate by comma if more than one.",
        )

        parser.add_argument(
            "--trial-prefix",
            dest="trial_prefix",
            default="",
            help="if specified, exports default models for a clinicedc trial",
        )

        parser.add_argument(
            "--skip_model",
            dest="skip_model_names",
            default="",
            help="models to skip in label_lower format. Separate by comma if more than one.",
        )

        parser.add_argument(
            "-p",
            "--path",
            dest="path",
            default=False,
            help="export path",
        )

        parser.add_argument(
            "-f",
            "--format",
            dest="format",
            default="csv",
            choices=["csv", "stata"],
            help="export format (csv, stata)",
        )

        parser.add_argument(
            "--stata-dta-version",
            dest="stata_dta_version",
            default=None,
            choices=["118", "119"],
            help="STATA DTA file format version",
        )

        parser.add_argument(
            "--include-historical",
            action="store_true",
            dest="include_historical",
            default=False,
            help="export historical tables",
        )

        parser.add_argument(
            "--decrypt",
            action="store_true",
            dest="decrypt",
            default=False,
            help="decrypt",
        )

        parser.add_argument(
            "--use-simple-filename",
            action="store_true",
            dest="use_simple_filename",
            default=False,
            help="do not use app_label or datestamp in filename",
        )

        parser.add_argument(
            "--country",
            dest="countries",
            default="",
            help="only export data for country. Separate by comma if more than one. ",
        )

        parser.add_argument(
            "--site",
            dest="site_ids",
            default="",
            help="only export data for site id. Separate by comma if more than one.",
        )

    def handle(self, *args, **options):  # noqa: ARG002
        user = get_export_user()
        validate_user_perms_or_raise(user, options["decrypt"])

        self.options = options
        self.decrypt = self.options["decrypt"]
        export_format = (
            CSV
            if self.options["format"] == "csv"
            else int(self.options["stata_dta_version"] or STATA_14)
        )

        if not self.options["path"] or not Path(self.options["path"]).expanduser().exists():
            raise CommandError(f"Path does not exist. Got `{self.options['path']}`")
        export_path = Path(self.options["path"]).expanduser().resolve()

        use_simple_filename = self.options["use_simple_filename"]

        sys.stdout.write("Export models.\n")
        sys.stdout.write(f"* export base path: {export_path}\n")

        if self.options["trial_prefix"]:
            model_names = get_default_models_for_export(self.options["trial_prefix"])
        else:
            app_labels = []
            if self.options["app_labels"]:
                app_labels = self.options["app_labels"].split(",")
                sys.stdout.write(
                    f"* preparing to export models from apps: {', '.join(app_labels)}\n"
                )
            model_names = []
            if self.options["model_names"]:
                model_names = self.options["model_names"].split(",")
                sys.stdout.write(f"* preparing to export models: {', '.join(model_names)}\n")
            if not app_labels and not model_names:
                raise CommandError(
                    "Nothing to do. No models to export. "
                    "Specify `app_labels` or `model_names`."
                )
            model_names = get_model_names_for_export(
                app_labels=app_labels,
                model_names=model_names,
            )

        if self.options["skip_model_names"]:
            skip_model_names = self.options["skip_model_names"].split(",")
            sys.stdout.write(f"* skipping models: {', '.join(skip_model_names)}\n")
            model_names = [m for m in model_names if m not in skip_model_names]

        if not self.options["include_historical"]:
            model_names = [m for m in model_names if "historical" not in m]

        # build list of site ids
        site_ids = self.options["site_ids"] or []
        if site_ids:
            site_ids = [int(x) for x in self.options["site_ids"].split(",")]
        site_ids = get_site_ids_for_export(site_ids=site_ids, countries=self.countries)

        # does user have perms to export these sites?
        for site_id in site_ids:
            site_sites.site_in_profile_or_raise(user, site_id)
        sys.stdout.write(
            f"* including data from sites: {', '.join([str(x) for x in site_ids])}\n\n"
        )

        if not model_names:
            raise CommandError("Nothing to do. No models to export.")

        # export
        models_to_file = ModelsToFile(
            user=user,
            models=model_names,
            site_ids=site_ids,
            decrypt=self.decrypt,
            archive_to_single_file=True,
            export_format=export_format,
            use_simple_filename=use_simple_filename,
            export_folder=export_path,
        )
        sys.stdout.write(
            style.SUCCESS(f"\nDone.\nExported to {models_to_file.archive_filename}\n")
        )

    @property
    def countries(self):
        if not self._countries:
            if not self.options["countries"] or self.options["countries"] == ALL_COUNTRIES:
                self._countries = site_sites.countries
            else:
                self._countries = self.options["countries"].lower().split(",")
                for country in self._countries:
                    if country not in site_sites.countries:
                        raise CommandError(f"Invalid country. Got {country}.")
        return self._countries



from .export_models_view import ExportModelsView
from .export_selected_models_view import ExportSelectedModelsView
from .home_view import HomeView

__all__ = ["ExportModelsView", "ExportSelectedModelsView", "HomeView"]



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_randomization.site_randomizers import site_randomizers
from edc_registration.models import RegisteredSubject
from edc_visit_schedule.model_mixins import VisitCodeFieldsModelMixin

from ...exceptions import StockRequestItemError
from ..medication import Assignment
from ..prescription import Rx
from .stock_request import StockRequest


class Manager(models.Manager):
    use_in_migrations = True


class StockRequestItem(VisitCodeFieldsModelMixin, BaseUuidModel):
    """A model that represents a stock request item."""

    request_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    request_item_datetime = models.DateTimeField(default=timezone.now)

    stock_request = models.ForeignKey(
        StockRequest,
        verbose_name="Stock request",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    rx = models.ForeignKey(Rx, on_delete=models.PROTECT, null=True, blank=False)

    assignment = models.ForeignKey(Assignment, on_delete=models.PROTECT, null=True, blank=True)

    registered_subject = models.ForeignKey(
        RegisteredSubject,
        verbose_name="Subject",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    appt_datetime = models.DateTimeField(null=True, blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return (
            f"{self.registered_subject.subject_identifier}: {self.stock_request.formulation}"
        )

    def save(self, *args, **kwargs):
        """Important: check `bulk_create_stock_request_items`
        to ensure fields updated here are also manually
        updated when `bulk_update` is called.
        """
        if not self.request_item_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.request_item_identifier = f"{next_id:06d}"
        if not self.stock_request:
            raise StockRequestItemError("Stock request may not be null")
        if self.registered_subject.subject_identifier != self.rx.subject_identifier:
            raise StockRequestItemError(
                "Subject mismatch. Selected subject does not match prescription"
            )
        self.assignment = self.rx.get_assignment()
        super().save(*args, **kwargs)

    @classmethod
    def get_rando_sid(cls, randomizer_name: str, registered_subject: RegisteredSubject):
        randomizer = site_randomizers.get(randomizer_name)
        rando_obj = randomizer.model_cls().objects.get(
            subject_identifier=registered_subject.subject_identifier
        )
        return rando_obj.sid

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock request item"
        verbose_name_plural = "Stock request items"



# Generated by Django 3.2.13 on 2022-09-13 21:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_registration", "0024_auto_20220704_1841"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalregisteredsubject",
            name="ethnicity",
            field=models.CharField(blank=True, max_length=15, null=True),
        ),
        migrations.AddField(
            model_name="registeredsubject",
            name="ethnicity",
            field=models.CharField(blank=True, max_length=15, null=True),
        ),
    ]



class EdcTemplateMiddlewareMixin:
    def check_for_required_request_attrs(self, request):
        try:
            request.url_name_data  # noqa: B018
        except AttributeError:
            request.url_name_data = {}
        try:
            request.template_data  # noqa: B018
        except AttributeError:
            request.template_data = {}



from datetime import date, datetime
from zoneinfo import ZoneInfo


def get_datetime_from_env(
    year: int,
    month: int,
    day: int,
    hour: int,
    minute: int,
    second: int,
    time_zone: str,
    closing_date: date | None = None,
) -> datetime:
    if closing_date:
        hour = hour or 23
        minute = minute or 59
        second = second or 59
    else:
        hour = hour or 0
        minute = minute or 0
        second = second or 0
    return datetime(
        int(year),
        int(month),
        int(day),
        int(hour),
        int(minute),
        int(second),
        0,
        ZoneInfo(time_zone),
    )



from django import forms

from edc_model_fields.widgets import SliderWidget


class MedicationAdherenceFormMixin:
    visual_score_slider = forms.CharField(
        label="Visual Score", widget=SliderWidget(attrs={"min": 0, "max": 100})
    )



from django.apps import apps as django_apps
from django.contrib import messages
from django.utils.translation import gettext as _

from edc_pharmacy.constants import CENTRAL_LOCATION


def may_delete_allocation(modeladmin, request, obj=None) -> bool:
    """Hack to block delete if the stock instance has been
    transferred to a location other than CENTRAL.

    Block as if the user does not have permissions to
    delete. An additional message shows why.

    User requires explicit permission to delete this model
    (which is not the default for any role).

    See AllocationAdmin
    """
    stock_transfer_item_model_cls = django_apps.get_model("edc_pharmacy.stocktransferitem")
    if obj and (
        (
            (
                stock_transfer_item := (
                    stock_transfer_item_model_cls.objects.filter(code=obj.stock.code)
                    .order_by("stock_transfer__transfer_datetime")
                    .last()
                )
            )
            and stock_transfer_item.stock.location.name != CENTRAL_LOCATION
        )
        or obj.stock.dispensed
    ):
        if stock_transfer_item:
            msg = _(
                "%s for stock %s can't be deleted because the stock item has already "
                "been transferred to another location. See stock transfer %s."
            ) % (
                obj._meta.object_name,
                obj.stock.code,
                stock_transfer_item.stock_transfer.transfer_identifier,
            )
        else:
            msg = _(
                "%s for stock %s can't be deleted because the stock item has "
                "already been dispensed."
            ) % (obj._meta.object_name, obj.stock.code)

        modeladmin.message_user(request, msg, messages.ERROR)
        return False
    return True



from django import forms

from ..models import BoxType


class BoxTypeForm(forms.ModelForm):
    def clean(self) -> dict:
        cleaned_data = super().clean()
        if cleaned_data.get("across") and cleaned_data.get("down"):
            if cleaned_data.get("across") * cleaned_data.get("down") != cleaned_data.get(
                "total"
            ):
                raise forms.ValidationError({"total": "Invalid total for given dimensions"})
        return cleaned_data

    class Meta:
        fields = "__all__"
        model = BoxType



import os

from django.core.checks import CheckMessage, Warning

from edc_export.utils import get_export_folder, get_upload_folder


def edc_export_checks(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    if not os.path.exists(get_export_folder()):
        errors.append(
            Warning(
                (
                    f"Folder does not exist. Tried {get_export_folder()}. "
                    "See settings.EDC_EXPORT_EXPORT_FOLDER."
                ),
                id="edc_export.W001",
            )
        )

    if not os.path.exists(get_upload_folder()):
        errors.append(
            Warning(
                (
                    f"Folder does not exist. Tried {get_upload_folder()}. "
                    "See settings.EDC_EXPORT_UPLOAD_FOLDER."
                ),
                id="edc_export.W002",
            )
        )

    return errors



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin

from .admin_site import edc_registration_admin
from .modeladmin_mixins import RegisteredSubjectModelAdminMixin
from .utils import get_registered_subject_model_cls


@admin.register(get_registered_subject_model_cls(), site=edc_registration_admin)
class RegisteredSubjectAdmin(
    SiteModelAdminMixin, RegisteredSubjectModelAdminMixin, SimpleHistoryAdmin
):
    ordering = ("subject_identifier",)

    fieldsets = (
        (
            "Subject",
            {
                "fields": (
                    "subject_identifier",
                    "sid",
                    "subject_type",
                    "registration_status",
                    "registration_datetime",
                )
            },
        ),
        (
            "Personal Details",
            {
                "fields": (
                    "first_name",
                    "last_name",
                    "initials",
                    "dob",
                    "gender",
                    "identity",
                )
            },
        ),
        (
            "Screening Details",
            {
                "fields": (
                    "screening_identifier",
                    "screening_datetime",
                )
            },
        ),
        (
            "Consent Details",
            {"fields": ("consent_datetime", "subject_consent_id")},
        ),
        (
            "Registration Details",
            {
                "fields": (
                    "randomization_list_model",
                    "randomization_datetime",
                    # "sid",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    fieldsets_no_pii = (
        (
            "Subject",
            {
                "fields": (
                    "subject_identifier",
                    "sid",
                    "subject_type",
                    "registration_status",
                    "registration_datetime",
                )
            },
        ),
        ("Personal Details", {"fields": ("gender",)}),
        (
            "Registration Details",
            {
                "fields": (
                    "screening_identifier",
                    "screening_datetime",
                    "randomization_datetime",
                    "consent_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )



from __future__ import annotations

from dataclasses import dataclass, field

from edc_model_fields.utils import Choices


@dataclass
class Row:
    choice: tuple[str, str] | tuple[str, str, str]
    extra: str | None = None

    name: str = field(init=False)
    display_name: str = field(init=False)
    custom_name: str | None = field(default=None, init=False)

    def __post_init__(self):
        try:
            self.name, self.display_name, self.custom_name = self.choice
        except ValueError:
            self.name, self.display_name = self.choice


@dataclass
class AsListData:
    data: tuple[tuple[str, str, str], ...] | Choices
    rows: list[Row] | None = field(default_factory=list, init=False)

    def __post_init__(self):
        for tpl in self.data:
            self.rows.append(Row(tpl))

    def __call__(self) -> tuple[Row, ...]:
        return tuple(self.rows)

    def __iter__(self):
        for row in self.rows:
            yield row.name, row.display_name, row.custom_name



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from django.db.models import QuerySet

from ..exceptions import BoundariesOverlap, NotEvaluated, ValueBoundryError
from .get_normal_data_or_raise import get_normal_data_or_raise
from .grading_data_model_cls import grading_data_model_cls

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from ..models import GradingData, NormalData, ReferenceRangeCollection

__all__ = ["get_grade_for_value"]


def get_grade_for_value(
    reference_range_collection: ReferenceRangeCollection,
    value: float | int | None = None,
    label: str | None = None,
    units: str | None = None,
    gender: str | None = None,
    dob: date | None = None,
    report_datetime: datetime | None = None,
    age_units: str | None = None,
    site: Site | None = None,
    create_missing_normal: bool | None = None,
) -> tuple[GradingData, str] | None:
    found_grading_data = None
    found_condition_str = None
    grading_data_objs = get_grading_data_instances(
        reference_range_collection=reference_range_collection,
        label=label,
        units=units,
        gender=gender,
        dob=dob,
        report_datetime=report_datetime,
        age_units=age_units,
    )
    for grading_data in grading_data_objs:
        normal_data = get_normal_data_or_raise(
            reference_range_collection=reference_range_collection,
            label=label,
            units=units,
            gender=gender,
            dob=dob,
            report_datetime=report_datetime,
            age_units=age_units,
            site=site,
            create_missing_normal=create_missing_normal,
        )
        lower_limit = get_lower_limit(normal_data, grading_data)
        upper_limit = get_upper_limit(normal_data, grading_data)
        value = float(value)
        condition_str = (
            f"{'' if lower_limit is None else lower_limit}"
            f"{grading_data.lower_operator or ''}{value}"
            f"{grading_data.upper_operator or ''}{'' if upper_limit is None else upper_limit}"
        )
        if eval(condition_str):  # nosec B307  # noqa: S307
            if not found_grading_data:
                found_grading_data = grading_data
                found_condition_str = (
                    f"{label}: {condition_str} {units} GRADE{grading_data.grade}"
                )
            else:
                raise BoundariesOverlap(
                    f"Overlapping grading definitions. Got {found_grading_data} "
                    f"which overlaps with {grading_data}. "
                    f"Using value={value} ({condition_str}). "
                    f"Check your grading definitions for `{label}` .",
                )
    return found_grading_data, found_condition_str


def get_lower_limit(
    normal_data: QuerySet[NormalData], grading_data: QuerySet[GradingData]
) -> int | float | None:
    lower_limit = float(grading_data.lower) if grading_data.lower else None
    if lower_limit and grading_data.lln:
        lower_limit = (
            lower_limit * normal_data.lower
            if "LLN" in grading_data.lln
            else lower_limit * normal_data.upper
        )
    return lower_limit


def get_upper_limit(
    normal_data: QuerySet[NormalData], grading_data: QuerySet[GradingData]
) -> int | float | None:
    upper_limit = float(grading_data.upper) if grading_data.upper else None
    if upper_limit and grading_data.uln:
        upper_limit = (
            upper_limit * normal_data.lower
            if "LLN" in grading_data.uln
            else upper_limit * normal_data.upper
        )
    return upper_limit


def get_grading_data_instances(
    reference_range_collection: ReferenceRangeCollection,
    label: str | None = None,
    units: str | None = None,
    gender: str | None = None,
    dob: date | None = None,
    report_datetime: datetime | None = None,
    age_units: str | None = None,
    site: Site | None = None,
) -> list[GradingData]:
    if not gender:
        raise ValueError("Gender may not be None")
    grading_data_objs = []
    qs = (
        grading_data_model_cls()
        .objects.filter(
            reference_range_collection=reference_range_collection,
            label=label,
            units=units,
            gender=gender,
            # site=site,
        )
        .order_by("grade")
    )
    for obj in qs.all():
        try:
            age_in_bounds = obj.age_in_bounds_or_raise(
                dob=dob,
                report_datetime=report_datetime,
                age_units=age_units,
            )
        except ValueBoundryError:
            pass
        else:
            if age_in_bounds:
                grading_data_objs.append(obj)
    if not grading_data_objs:
        if qs.count() == 0:
            msg = f"No matching grading data found for {label} {units} {gender}"
        else:
            msg = (
                f"No matching grading data found for {label} {units} {gender} given age bounds"
            )
        raise NotEvaluated(f"Value not graded. {msg}")
    return grading_data_objs



from edc_list_data.model_mixins import ListModelMixin


class RefusalReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Refusal Reason"
        verbose_name_plural = "Refusal Reasons"



# Generated by Django 6.0 on 2026-01-13 17:05

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0094_auto_20260113_0357"),
    ]

    operations = []



from django.contrib.sites.models import Site

from edc_model.models import HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin
from edc_timepoint.model_mixins import TimepointLookupModelMixin
from edc_timepoint.timepoint_lookup import TimepointLookup
from edc_visit_tracking.managers import CrfCurrentSiteManager, CrfModelManager

from .crf_no_manager_model_mixin import CrfNoManagerModelMixin


class CrfModelMixin(SiteModelMixin, TimepointLookupModelMixin, CrfNoManagerModelMixin):
    timepoint_lookup_cls = TimepointLookup

    objects = CrfModelManager()
    on_site = CrfCurrentSiteManager()
    history = HistoricalRecords(inherit=True)

    def get_site_on_create(self) -> Site:
        """Expect site instance to be set from the related_visit
        model instance.
        """
        return self.related_visit.site

    class Meta(CrfNoManagerModelMixin.Meta):
        abstract = True
        indexes = CrfNoManagerModelMixin.Meta.indexes



from .clinical_review_baseline_model_mixin import ClinicalReviewBaselineModelMixin
from .clinical_review_followup import (
    ClinicalReviewCholModelMixin,
    ClinicalReviewDmModelMixin,
    ClinicalReviewHivModelMixin,
    ClinicalReviewHtnModelMixin,
    ClinicalReviewModelMixin,
)
from .dx_location_model_mixin import DxLocationModelMixin
from .factory import (
    dx_initial_review_model_mixin_factory,
    rx_initial_review_model_mixin_factory,
)
from .followup_review import FollowupReviewModelMixin, HivFollowupReviewModelMixin
from .initial_review import (
    CholInitialReviewModelMixin,
    HivArvInitiationModelMixin,
    HivArvMonitoringModelMixin,
    NcdInitialReviewModelMixin,
)



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_reportable", "0006_alter_gradingdata_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="gradingdata",
            name="age_phrase",
            field=models.CharField(
                default="", help_text="calculated in save()", max_length=25
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="age_upper_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="description",
            field=models.CharField(default="", max_length=255),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="lln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="lower_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="phrase",
            field=models.CharField(
                default="",
                help_text="calculated by the formula instance",
                max_length=50,
                verbose_name="Value phrase",
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="reference_group",
            field=models.CharField(default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="uln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="upper_operator",
            field=models.CharField(default="", max_length=2),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="age_phrase",
            field=models.CharField(
                default="", help_text="calculated in save()", max_length=25
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="age_upper_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="description",
            field=models.CharField(default="", max_length=255),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="lln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="lower_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="phrase",
            field=models.CharField(
                default="",
                help_text="calculated by the formula instance",
                max_length=50,
                verbose_name="Value phrase",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="reference_group",
            field=models.CharField(default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="uln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="upper_operator",
            field=models.CharField(default="", max_length=2),
        ),
        migrations.AlterField(
            model_name="historicalmolecularweight",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalmolecularweight",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalmolecularweight",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmolecularweight",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="age_phrase",
            field=models.CharField(
                default="", help_text="calculated in save()", max_length=25
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="age_upper_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="description",
            field=models.CharField(default="", max_length=255),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="lln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="lower_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="phrase",
            field=models.CharField(
                default="",
                help_text="calculated by the formula instance",
                max_length=50,
                verbose_name="Value phrase",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="reference_group",
            field=models.CharField(default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="uln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="upper_operator",
            field=models.CharField(default="", max_length=2),
        ),
        migrations.AlterField(
            model_name="molecularweight",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="molecularweight",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="molecularweight",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="molecularweight",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_phrase",
            field=models.CharField(
                default="", help_text="calculated in save()", max_length=25
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_upper_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="description",
            field=models.CharField(default="", max_length=255),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="lln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="lower_operator",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="phrase",
            field=models.CharField(
                default="",
                help_text="calculated by the formula instance",
                max_length=50,
                verbose_name="Value phrase",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="reference_group",
            field=models.CharField(default="", max_length=25),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="uln",
            field=models.CharField(default="", max_length=15),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="upper_operator",
            field=models.CharField(default="", max_length=2),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
    ]
